from .base import BashTask, S3DownloadTask, DownloadTask, S3UploadTask, Task, SSHCommandTask, DepTask, TempDownloadTask, DownloadAnyTask, DockerContainerTask, MLDockerTask
from .base import LocalTarget, Target, S3Target, encode, encode_short, Uri
from .schedule import schedule, run, NoResultException, UnresolvedDependencyException
from .s3 import S3File, S3Obj, S3
from .ssh import get_client as get_ssh_client, download_file as ssh_download_file, upload_file as ssh_upload_file
from .utils import download, batching, get_files, parallize, logger



from typing import Callable
import copy
import logging
import io
from timeit import default_timer
from abc import ABC
from . import s3
from typing import List, Dict, Union, Optional, Tuple, runtime_checkable, Protocol
from pathlib import Path
from .utils import download
from . import ssh
import subprocess
import json
from PIL import Image
import numpy as np
import base64
import hashlib
from .docker import Container
import os
import shutil
from .utils import logger


class Uri:
    def __init__(self, uri: str) -> None:
        self.uri = uri

    def __repr__(self):
        return self.uri

    def __hash__(self) -> int:
        return hash(self.uri)

    def __eq__(self, __o: object) -> bool:
        return __o.uri == self.uri

    def __str__(self):
        return f"<Uri uri={self.uri}>"


def encode(url: str):
    return str(base64.urlsafe_b64encode(bytes(url, "utf-8")), 'utf-8')


def encode_short(url: str):
    e = hashlib.sha1(bytes(url, 'utf-8'))
    return e.hexdigest()


@runtime_checkable
class Comparable(Protocol):
    def _get_args(self):
        valid_items = filter(lambda x: not x[0].startswith("_"), self.__dict__.items())
        return ",".join(list(map(lambda x: F"{x[0]}={x[1]}", valid_items)))

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._get_args()}>"

    def __hash__(self) -> int:
        return hash(self.__repr__())

    def __eq__(self, __o: "Task") -> bool:
        return hash(__o) == hash(self)


class Target(Comparable):
    def exists(self) -> bool:
        """ checks whether or not the target exists """
        raise NotImplementedError()

    def delete(self):
        """ deletes the target """
        raise NotImplementedError()


class S3Target(s3.S3File):

    def delete(self):
        return super().unlink()

    @classmethod
    def from_uri(cls, uri: str):
        bucket, path = s3.S3.split_uri(uri)
        return S3Target(bucket, path)

    def __repr__(self) -> str:
        return


class LocalTarget(Target):
    def __init__(self, path: Union[str, Path], *args):
        self.path = Path(path, *args).absolute()

    def exists(self):
        return self.path.exists()

    def open(self, mode: str, **args):
        return self.path.open(mode=mode, **args)

    def delete(self):
        return self.path.unlink(missing_ok=True)

    def read_text(self):
        with self.open("r") as reader:
            data = reader.read()

        return data

    def write_text(self, text: str):
        with self.open("w") as writer:
            writer.write(text)

    def write_json(self, data: dict):

        with self.open('w') as writer:
            writer.write(json.dumps(data))

    def read_image(self, pil: bool = False):
        if pil:
            return Image.open(self.path)
        else:
            return np.asarray(Image.open(self.path))

    def read_json(self):
        return json.load(self.path.open('r'))

    def md5(self):
        hash_md5 = hashlib.md5()
        with self.path.open("rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)

        return hash_md5.hexdigest()

    def copy(self, destination: Union[str, Path, "LocalTarget"], exist_ok: bool = True):

        if isinstance(destination, LocalTarget):
            destination_path = destination.path
        else:
            destination_path = Path(destination)

        if destination_path.exists() and not exist_ok:
            raise FileExistsError(f"destination {destination} already exists. [Copy {self.path}]")

        if not self.path.exists():
            raise FileExistsError(f"file {self.path} already exists. [Copy to {destination_path}")

        shutil.copyfile(self.path, destination_path)

        if isinstance(destination, LocalTarget):
            return destination
        else:
            return LocalTarget(destination_path)


OutputType = Union[Target, List[Target], Dict[str, Target]]
Dependency = Union[Target, List[Target], "Task", List["Task"], Dict[str, Target], Dict[str, "Task"]]


def to_list(o: Optional[OutputType]):
    if o is None:
        return []

    elif isinstance(o, dict):
        return list(o.values())
    elif isinstance(o, list):
        return o
    elif isinstance(o, tuple):
        return list(o)
    else:
        return [o]


def depedendencies_resolved(deps: Dependency) -> bool:
    deps = to_list(deps)

    if len(deps) == 0:
        return True

    return all(o.exists() if isinstance(o, Target) else o.done() for o in deps)


class Task(Comparable):
    def depends(self) -> Dependency:
        return []

    def run(self):
        """ run the task. write to the target """
        raise NotImplementedError(f"task {self.__class__.__name__} does not implement run() method")

    def target(self) -> OutputType:
        """ task must not need a target, but task will always be exectued if the target is not defined"""
        return None

    def done(self):
        return all(o.exists() for o in to_list(self.target()))

    def runnable(self) -> bool:
        return depedendencies_resolved(self.depends())

    def unresolved_dependencies(self):
        for dep in to_list(self.depends()):
            if isinstance(dep, Task):
                if not dep.done():
                    yield dep
            elif isinstance(dep, LocalTarget):
                if not dep.exists():
                    yield dep

    def delete(self, recursive: bool = False):
        for t in to_list(self.target()):
            t.delete()

        if recursive:
            for dep in to_list(self.depends()):
                if isinstance(dep, Task):
                    dep.delete(recursive=recursive)

    def remote(self, ip: str, username: str, workdir: Union[str, Path], pem: Optional[Union[str, Path]] = None, executable: str = "python3"):
        client = ssh.get_client(ip, username, pem=pem)

        copy_task = copy.deepcopy(self)

        # sync dependencies
        for dep in to_list(copy_task.depends()):
            if isinstance(dep, Target):
                # sync local target to remote target
                targets = [dep]
            elif isinstance(dep, Task):
                targets = to_list(dep.target())
            else:
                targets = []

            for target in targets:
                if isinstance(target, LocalTarget):
                    if not target.exists():
                        raise Exception(f"could not find dependency {target}")

                    remote_target = LocalTarget(str(target.path.absolute()) + ".copy")
                    logger.debug(f"=> syncing target {target} to {remote_target}")
                    ssh.upload_file(client, target.path, remote_target.path)
                    target.path = remote_target.path

                elif isinstance(target, S3Target):
                    logger.debug(f"not syncing target {target}")

        # execute command python pickled task remotly

        # recreate environment
        # _, _, stderr = client.exec_command(f"ls {environment}")
        # environment_exists = len(stderr.readlines()) == 0

        # if not environment_exists:
        #     # create environment
        #     print("environment does not exist")

        # serialize object pickle
        path = f"/tmp/remote_task_{self.__class__.__name__}__{encode_short(self._get_args())}.pkl"
        remote_path = f"/tmp/task_{self.__class__.__name__}__{encode_short(copy_task._get_args())}.pkl"
        remote_result_path = remote_path + '.result'

        import pickle
        import os

        with open(path, 'wb') as writer:
            pickle.dump(copy_task, writer)

        assert (os.path.getsize(path) > 0)

        # upload serialized payload
        ssh.upload_file(client, path, remote_path)

        command = f"cd {workdir} && {executable} -c \"import pickle as p; task = p.load(open('{remote_path}', 'rb')); r=task.run(); p.dump(r,open('{remote_result_path}', 'wb'))\""
        logger.info("execute command: ", command)

        _, stdout, stderr = client.exec_command(command, get_pty=True, environment=os.environ)

        stderr = stderr.readlines()
        stdout = stdout.readlines()

        if len(stderr) > 0:
            raise Exception(f"failed to execute task remotely. stderr: {stderr}. {stdout}")

        # copy result from remote to local machine
        logger.debug("copy targts from remote to local")

        for target in to_list(self.target()):
            if isinstance(target, LocalTarget):
                local_path = target.path
                # local_path = str(target.path.absolute()) + ".remote"
                logger.info(f"download file {local_path}")
                ssh.download_file(client, target.path, local_path)

        # copy and read result
        local_result_path = remote_result_path + '.local'
        ssh.download_file(client, remote_result_path, local_result_path)

        return pickle.load(open(local_result_path, 'rb'))

    def _create_simple_local_target(self):
        args = self._get_args()
        return LocalTarget(f"/tmp/{self.__class__.__name__}_{encode_short(args)}.output")


class DepTask(Task):
    def run(self):
        pass

    def target(self) -> OutputType:
        return [dep.target() for dep in to_list(self.depends())]


class DownloadTask(Task):
    def __init__(self, url: str, destination: Path, auth: Optional[Tuple[str, str]] = None, headers: Optional[Dict[str, str]] = None) -> None:
        self.url: str = url
        self.destination: Path = Path(destination)
        self.headers = headers
        self.auth = auth

    def run(self):
        download(self.url, str(self.destination.absolute()), auth=self.auth, headers=self.headers)

    def target(self) -> LocalTarget:
        return LocalTarget(self.destination)


class TempDownloadTask(Task):
    def __init__(self, url: str, auth: Optional[Tuple[str, str]] = None, headers: Optional[Dict[str, str]] = None, suffix: Optional[str] = None) -> None:
        self.url: str = url

        filename = str(get_hash(url))
        if suffix:
            filename += suffix

        self.destination: Path = Path("/tmp/", filename).absolute()
        self.headers = headers
        self.auth = auth

    def run(self):
        logger.debug(f"downloading {self.url} to {self.destination}")
        start = default_timer()
        download(self.url, str(self.destination.absolute()), auth=self.auth, headers=self.headers)
        return dict(elapsed=default_timer() - start)

    def target(self) -> LocalTarget:
        return LocalTarget(self.destination)


def get_hash(obj: any) -> str:
    return encode_short(json.dumps(obj))


class BashTask(Task):
    def __init__(self, cmd: List[str]) -> None:
        self.cmd = cmd

    def run(self):
        logger.debug(f"[Bash] executing {self.cmd}")
        result = subprocess.Popen(self.cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        stderr = result.stderr.readlines()
        stdout = result.stdout.readlines()

        if (result.returncode is not None and result.returncode != 0) or len(stderr) > 0:
            raise Exception(f"task {self.__repr__()} [code={result.returncode}] has failed: {stderr}")

        logger.debug(f'stdout: {stdout}')
        logger.debug(f'stderr: {stderr}')

        with self.target().open('wb') as writer:
            writer.writelines(stdout)

        return stdout

    def target(self):
        return LocalTarget(f"/tmp/task_bash_{get_hash(self.cmd)}.output")


class SingleOutputTask(Task):

    def _run(self) -> Union[str, List[str]]:
        pass

    def run(self):
        output = self._run()

        with self.target().open("w") as writer:
            if isinstance(output, str):
                writer.write(output)
            else:
                writer.writelines(output)

        return output

    def target(self):
        args = self._get_args()
        return LocalTarget(f"/tmp/{self.__class__.__name__}_{encode_short(args)}.output")


class SSHCommandTask(Task):

    def __init__(self, ip: str, username: str, cmd: List[str], pem: Union[str, Path] = None) -> None:
        self.cmd = cmd
        self.pem = pem
        self.ip = ip
        self.username = username

    def run(self):

        logging.info(f"=> executing {self.cmd}")

        client = ssh.get_client(self.ip, self.username, pem=self.pem)
        stdin, stdout, stderr = client.exec_command(" ".join(self.cmd))

        stderr = stderr.readlines()

        if len(stderr) > 0:
            raise Exception(f"failed to execute {self.cmd} - stderr: {stderr}")

        out = stdout.readlines()

        with self.target().open('w') as writer:
            writer.writelines(out)

        return out

    def target(self):
        return self._create_simple_local_target()


class S3UploadTask(Task):

    def __init__(self, local_path: Union[str, Path], target_uri: Union[str, Tuple[str, str]]) -> None:
        self.local_path = Path(local_path)
        self.target_uri = target_uri if isinstance(target_uri, str) else f"s3://{target_uri[0]}/{target_uri[1]}"

    def depends(self) -> Dependency:
        return LocalTarget(self.local_path)

    def run(self):
        assert (self.target().upload(self.local_path))

    def target(self):
        return S3Target.from_uri(self.target_uri)


class S3DownloadTask(Task):

    def __init__(self, uri: Union[str, Tuple[str, str]], local_path: Union[str, Path]) -> None:
        self.local_path = Path(local_path)
        self.uri = uri if isinstance(uri, str) else f"s3://{uri[0]}/{uri[1]}"

    def depends(self):
        return S3Target.from_uri(self.uri)

    def run(self):
        self.depends().download(self.local_path)

    def target(self):
        return LocalTarget(str(self.local_path.absolute()))


class DockerContainerTask(SingleOutputTask):

    def __init__(self, name: str, force: bool = False, raise_exit_code_nonzero: bool = True) -> None:
        self.name = name
        self._container = Container.from_name(name)
        self.raise_exit_code_nonzero = raise_exit_code_nonzero

    def run(self):
        result = self._container.run(raise_exit_code_nonzero=self.raise_exit_code_nonzero)
        return result.json()


def _create_tmp_path(uri: str):
    ext = os.path.splitext(uri)[1]
    if "?" in ext:
        ext = ext.split("?")[0]
    ext = ext or ".output"
    return Path(f"/tmp/{encode_short(uri)}{ext}")


class CopyTask(Task):
    def __init__(self, source: Union[str, Path], destination: Union[str, Path]) -> None:
        self.source = source
        self.destination = destination

    def depends(self):
        return LocalTarget(self.source)

    def run(self):
        source = self.depends()
        target = self.target()

        if target.exists() and source.md5() == target.md5():
            logger.debug(f'copy {self.source} -> {self.destination} | target already exists with same hash')
            return dict(exists=True)
        else:
            source.copy(target)
            return dict(exists=False)

    def target(self):
        return LocalTarget(self.destination)


class DownloadAnyTask(DepTask):

    def __init__(self, uri: str, local_path: Union[str, Path] = None, **download_args):
        self.uri = uri
        self.download_args = download_args
        self.local_path = Path(local_path) if local_path else _create_tmp_path(uri)

    def depends(self):
        """ test uri format """
        if self.uri.startswith("s3://"):
            return S3DownloadTask(self.uri, self.local_path, **self.download_args)
        elif self.uri.startswith("http://") or self.uri.startswith("https://"):
            return DownloadTask(self.uri, self.local_path, **self.download_args)
        else:
            return CopyTask(self.uri, self.local_path)

    def target(self):
        return LocalTarget(self.local_path)


class LambdaTarget(Target):

    def __init__(self, lambda_fn: Callable):
        self.lambda_fn = lambda_fn

    def exists(self) -> bool:
        return self.lambda_fn()

    def delete(self):
        pass


class MLDockerTask(DockerContainerTask):

    def __init__(self, container_name: str, inputs: List[str], outputs: List[str], force: bool = False):
        self.inputs = inputs
        self.force = force
        self.container_name = container_name
        self._container = Container.from_name(container_name)

        # get input and output path
        logger.info(self._container.mounts)
        self.input_path = list(filter(lambda m: m.destination == Path('/input'), self._container.mounts))[0].source
        self.output_path = list(filter(lambda m: m.destination == Path('/output'), self._container.mounts))[0].source
        self.outputs = outputs

        # clear mounts
        def clear(dir: str):
            shutil.rmtree(dir, ignore_errors=True)
            os.makedirs(dir, exist_ok=True)

        clear(self.input_path)
        clear(self.output_path)

    def depends(self):
        return [
            DownloadAnyTask(uri, self.input_path / os.path.basename(uri)) for uri in self.inputs
        ] + [
            LambdaTarget(lambda: os.path.exists(self.output_path)),
            LambdaTarget(lambda: os.path.exists(self.input_path)),
        ]

    def run(self):
        return self._container.run(raise_exit_code_nonzero=True)

    def target(self):
        return [LocalTarget(self.output_path / o) for o in self.outputs]



from pathlib import Path
from timeit import default_timer
try:
    import docker
except:
    print("cannot import docker. DockerTasks will not not. Use <pip install docker>")

import enum
from pydantic import BaseModel
from typing import Optional
from datetime import datetime
from typing import Dict, List
import shutil
import os
from .utils import logger


class ContainerStatus(str, enum.Enum):
    exited = 'exited'
    running = 'running'
    paused = 'paused'
    restarting = 'restarting'
    invalid = 'invalid'
    dead = 'dead'
    created = 'created'


class Image(BaseModel):
    id: str
    name: str


class ContainerState(BaseModel):
    status: ContainerStatus
    exit_code: Optional[int]
    error: Optional[str]
    oom_killed: Optional[bool]
    pid: Optional[str]

    class Config:
        fields = {
            "exit_code": "ExitCode",
            "error": "Error",
            "status": "Status",
            "oom_killed": "OOMKilled",
            "pid": "Pid"
        }


class Mount(BaseModel):
    source: Path
    destination: Path
    mode: str
    rw: bool

    class Config:
        fields = {
            "source": "Source",
            "destination": "Destination",
            "mode": "Mode",
            "rw": "RW"
        }


class LastRun(BaseModel):
    start: datetime
    finish: datetime


class DockerException(Exception):
    pass


class ContainerAlreadyRunningException(DockerException):
    def __init__(self, container: "Container") -> None:
        self.container = container
        super().__init__(f"container {container.name} is already running")


class Result(BaseModel):
    logs: List[str]
    exit_code: int
    elapsed: float


class ContainerExitCodeException(DockerException):
    def __init__(self, container: "Container", exit_code: int,
                 logs: List[str],
                 elapsed: float):
        self.elapsed = elapsed
        self.exit_code = exit_code
        self.container = container
        self.logs = logs
        super().__init__(f"container {container.name} ran for {elapsed:.2f} seconds and has invalid exit code {exit_code}. Logs={logs}")


class ContainerNotFoundException(DockerException):
    pass


def parse_attrs(attrs: dict):

    container_id = attrs['Id']
    docker_image_id = attrs['Image']
    docker_image_name = attrs['Config']['Image']
    env = attrs['Config']['Env']
    state = attrs.get('State', {"Status": "invalid"})

    if state['Status'] == 'exited':
        last_start = state.get("StartedAt", None)
        last_finish = state.get("FinishedAt", None)
    else:
        last_start, last_finish = None, None
    env = dict([e.split("=") for e in env])
    restart_count = attrs.get("RestartCount", 0)
    mounts = [Mount(**m) for m in attrs.get("Mounts", [])]

    return dict(
        created=attrs.get("Created", None),
        restart_count=restart_count,
        env=env,
        id=container_id,
        state=ContainerState(
            **state
        ),
        image=Image(
            id=docker_image_id,
            name=docker_image_name
        ),
        last=None if last_start is None else LastRun(start=last_start, finish=last_finish),
        mounts=mounts
    )


class Container(BaseModel):
    name: str
    id: str
    created: datetime

    image: Image
    last: Optional[LastRun]

    state: ContainerState
    env: Dict[str, str]
    mounts: List[Mount]
    restart_count: int

    @classmethod
    def from_name(cls, name: str):
        obj = Container.obj(name)
        if obj is None:
            raise ContainerNotFoundException(f"container {name} cannot be found")

        attrs = Container.obj(name).attrs

        return cls(
            name=name,
            **parse_attrs(attrs)
        )

    @staticmethod
    def obj(name: str):
        return docker.from_env().containers.get(name)

    def get(self):
        return Container.obj(self.name)

    @property
    def attrs(self):
        return self.get().attrs

    def update(self):
        """ updates the attributes of this container """
        for k, v in parse_attrs(self.attrs).items():
            setattr(self, k, v)

    def run(self, raise_exit_code_nonzero: bool = False):
        self.update()

        if self.state.status == ContainerStatus.running:
            raise ContainerAlreadyRunningException(self)

        container = self.get()
        start_time = default_timer()

        logger.info(f'starting container {self.name}')
        container.start()

        logger.info(f"waiting for results for container {self.name}")
        result = container.wait()
        logger.info(f"container {self.name} finished: {result}")

        elapsed = default_timer() - start_time
        exit_code = result['StatusCode']

        logger.info(f"container {self.name} exit code: {exit_code}")

        logs = container.logs()
        logs = str(logs, 'utf-8')
        log_lines = logs.split('\n')

        if raise_exit_code_nonzero and exit_code != 0:
            raise ContainerExitCodeException(
                container=self,
                exit_code=exit_code,
                logs=log_lines,
                elapsed=elapsed
            )

        return Result(logs=log_lines, exit_code=exit_code, elapsed=elapsed)



import tqdm
import os
from typing import List, Tuple, Dict
import hashlib
import logging
import multiprocessing
import requests


def get_files(directory, extensions=None) -> List[str]:
    files = []
    extensions = [ext.lower() for ext in extensions] if extensions else None
    for root, _, filenames in sorted(os.walk(directory)):
        files += [os.path.join(root, name) for name in sorted(filenames)
                  if extensions is None or any([name.lower().endswith("." + ext) for ext in extensions])]

    return sorted(files)


def get_logger(name, level=logging.DEBUG, fh_level=logging.DEBUG, enable_file_logging=False, root_dir=''):

    logger = logging.getLogger(name)
    logger.propagate = False
    if not logger.handlers or len(logger.handlers) == 0:
        logger.setLevel(level)
        formatter = logging.Formatter(
            '%(asctime)s - %(levelname)s - %(message)s')
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG)
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

    if enable_file_logging and not any([isinstance(handler, logging.FileHandler) for handler in logger.handlers]):
        formatter = logging.Formatter(
            '%(asctime)s - %(levelname)s - %(message)s')
        fh = logging.FileHandler(os.path.join(root_dir, '%s.log' % name))
        fh.setLevel(fh_level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    return logger


def batching(args, batch_size: int):
    for i in range(0, len(args), batch_size):
        yield args[i:(i + batch_size)]


def parallize(f, args, n_processes: int = None, desc: str = 'parallize_v3', tq=None, postfix={}, batch_size: int = 64):
    if n_processes == None:
        n_processes = multiprocessing.cpu_count()

    with multiprocessing.Pool(n_processes) as pool:
        total = len(args)
        results = []

        if total == 0:
            return []

        if (type(args[0]) == list or type(args[0]) == tuple) and len(args[0]) > 1:
            desc = '[batching] ' + desc
            results = []

            if tq is None:
                with tqdm.tqdm(total=total, desc=desc) as tq:
                    for batch in batching(args, batch_size):
                        results.extend(pool.starmap(f, batch))
                        tq.update(len(batch))

            else:
                for batch in batching(args, batch_size):
                    results.extend(pool.starmap(f, batch))
                    tq.set_postfix(done="%d/%d" % (len(results) + 1, total), desc=desc, **postfix)

        else:
            if tq is None:
                for r in tqdm.tqdm(
                        pool.imap(f, args), desc=desc, total=total):
                    results.append(r)
            else:
                for i, r in enumerate(pool.imap(f, args)):
                    tq.set_postfix(done="%d/%d" % (i + 1, total), desc=desc, **postfix)
                    results.append(r)

    return results


logger = get_logger("task_manager")


def download(url: str, destination: str, auth: Tuple = None, headers: Dict[str, str] = None):
    r = requests.get(url, auth=auth, headers=headers)
    if r.status_code != 200:
        logger.error("url: %s | status code: %d" % (url, r.status_code))
        raise ValueError("status code == %d" % r.status_code)

    with open(destination, 'wb') as w:
        w.write(r.content)


def encode(data: str):
    m = hashlib.sha256()
    m.update(bytes(data, 'utf-8'))
    return m.hexdigest()



import codecs
from collections import namedtuple
import io
from pathlib import Path
import tqdm
import os
from dataclasses import dataclass
import logging
from typing import Any, Dict, List, Tuple, Union, Optional
import boto3
import botocore
import io
from PIL import Image
import numpy as np
from typing import Protocol
import functools
import ujson as json
import pickle
from timeit import default_timer
from .utils import logger

get_var = os.getenv


class StreamingBody(Protocol):
    def read(self) -> bytes:
        pass


def seekable_stream(stream: StreamingBody):
    # StreamingBody
    return io.BytesIO(stream.read())


class S3ReadWrite:

    def __init__(self, s3file: "S3File", mode: str, show_tq: bool = False, seekable: bool = False):
        """ Stream needs to be seekable for reading numpy arrays"""

        self.file = s3file
        self.mode = mode
        self.show_tq = show_tq
        self.seekable = seekable

        if mode == 'r':
            byte_stream = S3().get_file_stream(self.file.path, self.file.bucket)
            StreamReader = codecs.getreader('utf-8')
            self.stream = StreamReader(byte_stream)

        elif mode == 'rb':
            self.stream = S3().get_file_stream(self.file.path, self.file.bucket)
            if self.seekable:
                self.stream = seekable_stream(self.stream)

        elif mode == 'wb':
            self.stream = io.BytesIO()
        elif mode == 'w':
            self.stream = io.StringIO()
        else:
            raise Exception(f"mode {mode} not supported")

    def __enter__(self):
        return self.stream

    def __exit__(self, type, value, traceback):
        # persist the stream
        if self.mode == 'w' or self.mode == 'wb':
            num_bytes = self.stream.tell()
            self.stream.seek(0)

            if isinstance(self.stream, io.StringIO):
                data = self.stream.getvalue()
                bytestream = io.BytesIO(bytes(data, 'utf-8'))
                S3().write_stream(bytestream, self.file.path, self.file.bucket, num_bytes=num_bytes, show_tq=self.show_tq)
            else:
                S3().write_stream(self.stream, self.file.path, self.file.bucket, num_bytes=num_bytes, show_tq=self.show_tq)

        self.stream.close()

    def read(self, size: int = -1) -> bytes:
        if not hasattr(self.stream, "read"):
            self.stream = seekable_stream(self.stream)
        else:
            return self.stream.read(size)


S3Obj = namedtuple('S3Obj', ['key', 'mtime', 'size', 'ETag'])


@dataclass
class S3File:
    bucket: str
    path: str

    def upload(self, local_path: Union[str, Path], progress: bool = False) -> bool:
        if isinstance(local_path, Path):
            local_path = str(local_path.absolute())

        S3().upload_file(local_path, self.path, self.bucket, progress=progress)
        return self.exists()

    def download(self, path: Union[str, Path]):
        if isinstance(path, Path):
            path = str(path.absolute())

        S3().download_file(self.path, path, self.bucket)

    def unlink(self):
        return self.rmtree()

    @property
    def size(self) -> Optional[float]:
        """ File size via HEAD method. Might return invalid size incase of encryption
            returns: float (bytes) or None, if path is not found
        """
        response = S3().s3_client.head_object(Bucket=self.bucket, Key=self.path)
        if "ContentLength" not in response:
            return None
        return response['ContentLength']

    @property
    def parent(self):
        from pathlib import Path
        return S3File(path=str(Path(self.path).parent), bucket=self.bucket)

    @property
    def uri(self):
        return "s3://%s/%s" % (self.bucket, self.path)

    @classmethod
    def from_uri(cls, uri: str):
        bucket, path = S3.split_uri(uri)
        return S3File(bucket, path)

    def exists(self):
        return S3().file_exists(self.path, self.bucket) or S3().directory_exists(self.path, self.bucket)

    def open(self, mode: str, show_tq: bool = False, seekable: bool = False):
        return S3ReadWrite(self, mode, show_tq, seekable=seekable)

    def walk(self) -> List["S3File"]:
        paths = S3().list_directory(self.path, self.bucket)
        return list(map(lambda path: S3File(self.bucket, path), paths))

    def listdir(self, limit: int = -1):
        return list(map(lambda x: os.path.basename(x[:-1]), filter(lambda x: x.endswith("/"), self.list_prefixes(limit=limit))))

    def list_prefixes(self, limit: int = -1):
        return list(S3().list_prefixes_v2(self.path, self.bucket, limit=limit))

    def _head(self):
        return S3().s3_client.head_object(Bucket=self.bucket, Key=self.path)

    def content_length(self) -> float:
        return self._head()['ContentLength']

    def join(self, *args):
        f = self
        for a in args:
            f = f / a
        return f

    def __len__(self):
        return self.content_length()

    def __truediv__(self, other: str):
        new_path = os.path.join(self.path, other)
        return S3File(self.bucket, new_path)

    def rmtree(self):
        if not self.exists():
            return False

        S3().delete_directory(self.path, self.bucket)
        return True

    def mkdir(self, *args, **kwargs):
        """ ghost method because s3 directories don't need to be created """
        pass

    def read(self) -> bytes:
        """ make it behave like a stream """
        with self.open('rb') as reader:
            data = reader.read()

        return data

    def write_image(self, image: Union[np.ndarray, Image.Image], format: str = "JPEG", show_tq: bool = False):
        S3().write_image(image, self.path, self.bucket, format=format, show_tq=show_tq)

    def read_image(self, pil: bool = False) -> Union[np.ndarray, Image.Image]:
        return S3().read_image(self.path, self.bucket, pil=pil)

    def read_numpy(self):
        compressed = self.path.lower().endswith(".npz")

        with self.open("rb", seekable=True) as stream:
            if compressed:
                descriptor = np.load(stream)
                data = descriptor['data']
                descriptor.close()
            else:
                data = np.load(stream)

        return data

    def __check_numpy_file_ending(self, compressed: bool):
        if compressed:
            assert (self.path.endswith(".npz")), f"{self.path} has to end with .npz"
        else:
            assert (self.path.endswith(".npy")), f"{self.path} has to end with .npy"

    def write_arrays(self, data: Dict[str, np.ndarray], compressed: bool = True, show_tq: bool = False):
        self.__check_numpy_file_ending(compressed)
        with self.open('wb', show_tq=show_tq) as writer:
            if compressed:
                np.savez_compressed(writer, **data)
            else:
                np.save(writer, data)

    def read_arrays(self, compressed: bool = True, show_tq: bool = False) -> Dict[str, np.ndarray]:
        with self.open("rb", show_tq=show_tq, seekable=True) as stream:
            if compressed:
                data = np.load(stream, allow_pickle=True)
            else:
                data = np.load(stream, allow_pickle=True)

            if isinstance(data, np.ndarray):
                return data.tolist()  # returns dict :D
            else:
                keys = list(data.keys())
                ret = {d: data[d] for d in keys}
                return ret

    def write_numpy(self, data: np.ndarray, compressed: bool = True, show_tq: bool = False):
        # ensure correct line ending
        self.__check_numpy_file_ending(compressed)

        with self.open('wb', show_tq=show_tq) as writer:
            if compressed:
                np.savez_compressed(writer, data=data)
            else:
                np.save(writer, data)

    def read_text(self, encoding='utf-8'):
        return str(self.read(), encoding)

    def write_text(self, text: str):
        with self.open("w") as writer:
            writer.write(text)

    def copy(self, dst: "S3File", check_exists: bool = False) -> Optional["S3File"]:
        if S3().copy(self.bucket, self.path, dst.path, dst.bucket, check_exists=check_exists):
            return dst
        else:
            return None

    def write_json(self, json_data: dict):
        data = bytes(json.dumps(json_data), 'utf-8')
        S3().write_bytes(data, self.path, self.bucket)

    def read_json(self) -> dict:
        with self.open("r") as reader:
            data = reader.read()

        return json.loads(data)

    def read_dict(self) -> dict:
        with self.open("r") as reader:
            data = reader.read()
        return json.loads(data)

    """ csv """

    def read_csv(self, **kwargs):
        import pandas as pd
        with self.open("rb") as reader:
            df = pd.read_csv(reader, **kwargs)
        return df

    def write_csv(self, df, index: bool = False, **kwargs):
        import pandas as pd
        with self.open("wb") as writer:
            df.to_csv(writer, index=index, **kwargs)
        return df

    """ pickle """

    def read_pickle(self, **kwargs) -> Any:
        with self.open("rb", seekable=True) as reader:
            data = pickle.load(reader, **kwargs)
        return data

    def write_pickle(self, obj: Any) -> Any:
        with self.open("wb", seekable=True) as writer:
            data = pickle.dump(obj, writer)
        return data

    """ publicity functions"""

    def get_public_url(self):
        location = S3().s3_client.get_bucket_location(Bucket=self.bucket)['LocationConstraint']
        return "https://s3-%s.amazonaws.com/%s/%s" % (location, self.bucket, self.path)

    def make_public(self) -> bool:
        response = S3().s3_client.put_object_acl(
            ACL="public-read", Bucket=self.bucket, Key=self.path
        )
        return response['ResponseMetadata']['HTTPStatusCode'] == 200

    def is_publicly_accessible(self):
        grants = self._get_grants()

        def has_read_rights(g):
            grantee = g['Grantee']
            permission = g['Permission']

            if permission == "READ" and "URI" in grantee and grantee['URI'] == "http://acs.amazonaws.com/groups/global/AllUsers":
                return True

            return False

        return any(has_read_rights(g) for g in grants)

    def _get_grants(self):
        return S3().s3_client.get_object_acl(
            Bucket=self.bucket,
            Key=self.path,
        )['Grants']

    def meta(self) -> Optional[S3Obj]:

        results = list(S3().walk_v2(self.path, self.bucket, limit=1))
        if len(results) == 0:
            return None
        r: S3Obj = results[0]
        if r.key != self.path:
            return None

        return r


def get_config():
    s3_region = get_var("S3_REGION")
    aws_access_key_id = get_var("ACCESS")
    aws_secret_access_key = get_var("SECRET")
    branch = get_var("APP_BRANCH")
    bucket = get_var("S3_BUCKET")

    return {
        "aws_secret_access_key": aws_secret_access_key,
        "s3_region": s3_region,
        "aws_access_key_id": aws_access_key_id,
        "s3_bucket": "%s-%s" % (bucket, branch.lower()),
        "profile": os.environ.get("AWS_PROFILE", "")
    }


class S3(object):
    _instace: Optional["S3"] = None
    MAX_SESSION_LENGTH = 60 * 60 * 11

    def __init__(self, *args, **kwargs):
        self.config = get_config()

        # check if init of this class was already called once
        has_session = hasattr(self, "start") and hasattr(self, "s3")

        if has_session and not self._session_exired():
            # logger.debug("reusing active session")
            pass
        else:
            self.invalidate_session()

    def invalidate_session(self):
        """ invalidates current session and recreates necessary properties of the class"""

        args = {}

        if self.config['profile'] == "":
            access_key_id = self.config['aws_access_key_id']
            logger.debug(f"start s3 session using access key {access_key_id}")
            session = boto3
            args = dict(
                region_name=self.config['s3_region'],
                aws_access_key_id=access_key_id,
                aws_secret_access_key=self.config['aws_secret_access_key']
            )
        else:
            profile = self.config['profile']
            logger.debug(f"start s3 session using profile {profile}")
            session = boto3.session.Session(profile_name=profile)

        self.s3_client = session.client('s3', **args)
        self.s3 = session.resource('s3', **args)
        self.transfer = boto3.s3.transfer.S3Transfer(client=self.s3_client)

        # set the start of the new session
        self.start: float = default_timer()
        print("start is defined...")

    def _session_exired(self):
        total_seconds = default_timer() - self.start
        return total_seconds > self.MAX_SESSION_LENGTH

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)

        return cls._instance

    @staticmethod
    def split_uri(uri: str) -> Tuple[str, str]:
        # format: s3://my-bucket/my-path
        assert (uri.startswith("s3://")), "uri has to start with s3://"
        start = uri[5:]
        split = start.split("/")
        bucket = split[0]
        path = "/".join(split[1:])
        return bucket, path

    @staticmethod
    def make_uri(bucket: str, path: str) -> str:
        return f"s3://{bucket}/{path}"

    @staticmethod
    def join_uri(uri: str, filename: str) -> str:
        s3_bucket, s3_path = S3.split_uri(uri)
        return S3.make_uri(s3_bucket, os.path.join(s3_path, filename))

    @property
    def default_bucket(self):
        return self.config['s3_bucket']

    @property
    def default_region(self):
        return self.config['s3_region']

    def __init__(self):
        self.config = get_config()
        args = {}
        if self.config['profile'] == "":
            session = boto3
            args = dict(
                region_name=self.config['s3_region'],
                aws_access_key_id=self.config['aws_access_key_id'],
                aws_secret_access_key=self.config['aws_secret_access_key']
            )
        else:
            profile = self.config['profile']
            session = boto3.session.Session(profile_name=profile)

        self.s3_client = session.client('s3', **args)
        self.s3 = session.resource('s3', **args)
        self.transfer = boto3.s3.transfer.S3Transfer(client=self.s3_client)

    def copy(self, src_bucket: str, src_key: str, dst_key: str, dst_bucket: Optional[str] = None, check_exists: bool = False) -> bool:
        # returns whether or not the file was copied

        if dst_bucket == None:
            dst_bucket = src_bucket

        if check_exists and self.file_exists(dst_key, dst_bucket):
            return False

        return self.s3.Object(dst_bucket, dst_key).copy_from(CopySource={"Bucket": src_bucket, "Key": src_key})

    def file_exists(self, s3_path, bucket_name):
        # directory exists works as well for files
        return self.directory_exists(s3_path, bucket_name)

    def file_exists_head(self, s3_path: str, bucket_name: str):
        try:
            self.s3_client.head_object(Bucket=bucket_name, Key=s3_path)
            return True
        except botocore.errorfactory.ClientError as c:
            if "Error" in c.response and "Code" in c.response["Error"]:
                if c.response['Error']["Code"] == "403":
                    logger.error(f"403 Forbidden on s3://{bucket_name}/{s3_path}")

            return False

    def list_prefixes(self, s3_path: str, s3_bucket: str, limit: int = 1000) -> List[str]:
        prefix = s3_path[1:] if s3_path.startswith("/") else s3_path
        if not prefix.endswith("/"):
            prefix = prefix + "/"

        result = self.s3_client.list_objects(Bucket=s3_bucket, Delimiter="/", Prefix=prefix, MaxKeys=limit)
        result_prefixes = result.get("CommonPrefixes")

        if result_prefixes == None:
            return []

        return [o.get("Prefix") for o in result_prefixes]

    def list_prefixes_v2(self, s3_path: str, s3_bucket: str, limit: int = -1):
        prefix = s3_path[1:] if s3_path.startswith("/") else s3_path
        if not prefix.endswith("/"):
            prefix = prefix + "/"

        bucket = self.s3.Bucket(s3_bucket)
        paginator = bucket.meta.client.get_paginator('list_objects')
        num = 0
        args = dict(Bucket=bucket.name, Delimiter="/", Prefix=prefix)
        if limit > 0 and limit < 1000:
            args['MaxKeys'] = limit

        for result in paginator.paginate(**args):
            result_prefixes = result.get("CommonPrefixes")

            if result_prefixes == None:
                return []

            for r in result_prefixes:
                num += 1
                yield r.get("Prefix")

                if limit > 0 and num == limit:
                    break

            if limit > 0 and num == limit:
                break

    def directory_exists(self, s3_path, bucket_name) -> bool:

        try:
            result = self.s3_client.list_objects(Bucket=bucket_name, Prefix=s3_path, MaxKeys=1)
            if "Contents" in result:
                return True
            else:
                return False
        except botocore.errorfactory.ClientError as c:
            return False

    def upload_file(self, local_path, s3_path, bucket_name, progress: bool = False):
        if progress:
            with open(local_path, "rb") as reader:
                self._write_stream_with_progress(reader, s3_path, bucket_name, num_bytes=os.path.getsize(local_path), show_tq=True)
        else:
            self.transfer.upload_file(local_path, bucket_name, s3_path)

        logger.debug('S3: uploaded ' + local_path + ' to ' + s3_path)

    def download_file(self, s3_path, local_path, bucket_name, debug=True):
        """
        dir_parts = local_path.split('/')[:-1]
        act_dir = '/'
        for i in range(len(dir_parts)):
            act_dir = os.path.join(act_dir, dir_parts[i])
            if not os.path.exists(act_dir):
                os.makedirs(act_dir)
        """

        self.transfer.download_file(bucket_name, s3_path, local_path)
        if debug:
            logger.debug('S3: downloaded %s in bucket %s to %s' % (s3_path, bucket_name, local_path))

    def list_directory(self, s3_directory, bucket_name):
        """Get a list of all keys in an S3 bucket."""
        keys = []
        kwargs = {'Bucket': bucket_name, 'Prefix': s3_directory}
        while True:
            resp = self.s3_client.list_objects_v2(**kwargs)
            if "Contents" not in resp:
                break
            for obj in resp['Contents']:
                keys.append(obj['Key'])
            try:
                kwargs['ContinuationToken'] = resp['NextContinuationToken']
            except KeyError:
                break
        return keys

    def walk(self, s3_path: str, bucket_name: str):
        kwargs = {'Bucket': bucket_name, 'Prefix': s3_path}

        keys = []
        while True:
            resp = self.s3_client.list_objects_v2(**kwargs)
            if "Contents" not in resp:
                break

            for obj in resp['Contents']:
                keys.append(obj['Key'])
                # yield S3Obj(f['Key'], f['LastModified'], f['Size'], f['ETag'])
            try:
                kwargs['ContinuationToken'] = resp['NextContinuationToken']
            except KeyError:
                break
        return keys

    def walk_v2(self, s3_directory: str, bucket_name: str, limit: int = -1):
        # [S3Obj(f['Key'], f['LastModified'], f['Size'], f['ETag']) for f in resp['Contents']]
        kwargs = {'Bucket': bucket_name, 'Prefix': s3_directory}
        if limit > 0 and limit < 1000:
            kwargs['MaxKeys'] = limit

        num = 0

        while limit < 0 or num < limit:

            resp = self.s3_client.list_objects_v2(**kwargs)
            if "Contents" not in resp:
                break

            for f in resp['Contents']:
                num += 1
                yield S3Obj(f['Key'], f['LastModified'], f['Size'], f['ETag'])
                if limit > 0 and num == limit:
                    break

            try:
                kwargs['ContinuationToken'] = resp['NextContinuationToken']
            except KeyError:
                break

    def download_directory(self, s3_directory, local_directory, bucket_name):
        s3_list = self.list_directory(s3_directory, bucket_name)

        with tqdm.tqdm(total=len(s3_list), desc='downloading files from s3') as tq:
            for s3_url in s3_list:
                s3_dir = os.path.dirname(s3_url).replace(s3_directory, '')
                if len(s3_dir) > 0 and s3_dir[0] == '/':
                    s3_dir = s3_dir[1:]
                local_dir = os.path.join(local_directory, s3_dir)
                if not os.path.isdir(local_dir):
                    os.makedirs(local_dir)
                local_fn = os.path.join(local_dir, s3_url.split('/')[-1])
                try:
                    self.download_file(s3_url, local_fn, bucket_name, debug=False)
                    tq.set_postfix(file=s3_url.encode('ascii', "ignore"))
                except Exception as e:
                    logger.error('[S3] not downloading %s to %s (bucket=%s)' % (s3_url, local_fn, bucket_name))
                tq.update(1)

    def read_image(self, s3_path, bucket_name, pil: bool = False) -> np.ndarray:
        from PIL import Image

        obj = self.s3.Bucket(bucket_name).Object(s3_path)
        response = obj.get()
        stream = response['Body']
        image = Image.open(stream)
        if pil:
            return image

        return np.array(image)

    def read_bytes(self, s3_path, bucket_name):
        obj = self.s3.Bucket(bucket_name).Object(s3_path)
        response = obj.get()
        stream = response['Body']
        return stream.read()

    def get_file_stream(self, s3_path, bucket_name):
        obj = self.s3.Bucket(bucket_name).Object(s3_path)
        response = obj.get()
        stream = response['Body']
        return stream

    def write_bytes(self, bytes, s3_path, bucket_name, show_tq: bool = False):
        return self._write_stream_with_progress(io.BytesIO(bytes), s3_path, bucket_name, num_bytes=len(bytes), show_tq=show_tq)

    def _write_progress(self, num_bytes: int, tq):
        tq.update(num_bytes)

    def _write_stream_with_progress(self, stream, s3_path: str, bucket_name: str, num_bytes: int = None, show_tq: bool = True):
        total = None if num_bytes == None else num_bytes
        bar_format = "{desc}: {percentage:.1f}%|{bar}| {n:.2f}/{total:.2f} [{elapsed}<{remaining}]"
        unit_scale = 1. / (1024 * 1024)

        if show_tq and total != None:
            with tqdm.tqdm(desc=f"uploading {self.make_uri(bucket_name, s3_path)}", unit='MB', bar_format=bar_format, total=total, unit_scale=unit_scale) as tq:
                return self.s3.Bucket(bucket_name).upload_fileobj(stream, s3_path, Callback=lambda x, tq=tq: self._write_progress(x, tq))
        else:
            return self.s3.Bucket(bucket_name).upload_fileobj(stream, s3_path)

    def write_stream(self, stream, s3_path, bucket_name, num_bytes: int = None, show_tq: bool = True):
        return self._write_stream_with_progress(stream, s3_path, bucket_name, num_bytes=num_bytes, show_tq=show_tq)

    def write_url(self, file_url: str, s3_path: str, bucket: str, show_tq: bool = False):
        # writes url stream to s3
        import requests

        with requests.get(file_url, stream=True) as r:
            content_length = r.headers.get('Content-Length', None)
            S3().write_stream(r.raw, s3_path, bucket, show_tq=False)

    def write_image(self, image: Union[np.ndarray, Image.Image], s3_path, bucket_name, format='PNG', show_tq: bool = False):
        stream = io.BytesIO()

        if not isinstance(image, Image.Image):
            image = Image.fromarray(image)
        else:
            if image.format != None:
                logger.debug(f"using pil image format {image.format}")
                format = image.format

        image.save(stream, format=format, quality=100)

        # get the number of bytes to write by the current stream position
        num_bytes = stream.tell()

        # seek to the start to write the whole file
        stream.seek(0)

        self.write_stream(stream, s3_path, bucket_name, num_bytes=num_bytes, show_tq=show_tq)

    def delete_directory(self, s3_directory, bucket_name):
        s3_list = self.list_directory(s3_directory, bucket_name)
        with tqdm.tqdm(total=len(s3_list), desc='deleting files from s3') as tq:
            for path in s3_list:
                self.delete_file(path, bucket_name, debug=False)
                tq.set_postfix(file=path.encode('ascii', "ignore"))
                tq.update(1)

    def delete_file(self, s3_path, bucket_name, debug=True) -> bool:
        try:
            self.s3.Object(bucket_name, s3_path).delete()
            if debug:
                logger.debug("S3: deleted %s from bucket %s" % (s3_path, bucket_name))
            return True
        except Exception as e:
            return False

    def create_bucket(self, bucket_name):
        """ https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.create_bucket """

        self.s3_client.create_bucket(Bucket=bucket_name, CreateBucketConfiguration={
                                     'LocationConstraint': 'eu-central-1'})

    def guess_media_type(self, ext) -> Optional[str]:
        m = {
            "png": "image/png",
            "jpg": "image/jpeg",
            "jpeg": "image/jpeg",
            "pdf": "application/pdf",
            "xml": "application/xml",
            "zip": "application/zip",
            "ppt": "application/vnd.ms-powerpoint",
            "xls": "application/vnd.ms-excel",
            "doc": "application/msword",
            "pptx": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
            "xlsx": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
            "odt": "application/vnd.oasis.opendocument.text",
            "csv": "text/csv",
            "html": "text/html",
            "txt": "text/plain",
            "dxf": "image/vnd.dxf",
            "svg": "image/svg+xml",
            "json": "application/json"
        }
        if ext in m:
            return m[ext]
        else:
            return None


if __name__ == '__main__':
    _test()



from typing import Dict
import networkx as nx
from .base import Task, to_list, Target
from typing import List
from multiprocessing import Pool as ProcessPool
from multiprocessing.pool import ThreadPool
from datetime import datetime


def _schedule(task: Task, g: nx.Graph, force: bool = False):

    if not task.done() or force:
        g.add_node(task)

        for dep_task in to_list(task.depends()):

            if isinstance(dep_task, Task):
                g.add_edge(task, dep_task)
                g = _schedule(dep_task, g, force=force)

            elif isinstance(dep_task, Target):
                if not dep_task.exists():
                    raise UnresolvedDependencyException(task, unresolved=dep_task)
            else:
                raise Exception(f"invalid task/target type: {dep_task.__class__.__name__}")

    return g


def schedule(task: Task, force: bool = False) -> nx.Graph:
    return _schedule(task, nx.Graph(), force=force)


def get_path_lengths(root: Task, g: nx.Graph):
    for node in g.nodes:
        yield node, len(nx.shortest_path(g, root, node))


def create_execution_order(root: Task, g: nx.Graph) -> List[List[Task]]:
    # put task into bins
    levels: List[List[Task]] = []

    current_degree: int = 1

    stage: List[Task] = []

    for task, degree in sorted(list(get_path_lengths(root, g)), key=lambda x: x[1]):
        # print(f"adding task {task} to level={degree}")
        if degree != current_degree:
            levels.append(stage)
            stage = [task]
            current_degree = degree
        else:
            stage.append(task)

    levels.append(stage)
    levels = list(filter(lambda x: len(x) != 0, levels))
    return list(reversed(levels))


class UnresolvedDependencyException(Exception):
    def __init__(self, task: Task, unresolved: any = None) -> None:
        self.task = task
        self.unresolved = unresolved
        super().__init__(f"[{task}] has unfullfilled dependency {unresolved if unresolved else list(task.unresolved_dependencies())} for task {task}")


class NoResultException(Exception):
    def __init__(self, task: Task, target: Target) -> None:
        self.task = task
        super().__init__(f"[{task}] has not procuced target {target}")


def run(task: Task, PoolClass=ThreadPool, workers: int = 4, debug: bool = False, force: bool = False) -> Dict[Task, any]:
    g = schedule(task, force=force)
    order = create_execution_order(task, g)

    if debug:
        print("\n\n========= execution order =========")
        for level, tasks in enumerate(order):
            print(f"[level={level}] => ", tasks)
        print("===================================\n\n")

    start = datetime.now()
    print(f'start => {start}')

    all_results = {}

    with PoolClass(workers) as pool:
        for tasks in order:
            print(".")
            scheduled = []
            for task in tasks:
                print(f"=> Running {task}")

                if task.runnable():
                    scheduled.append(task)
                else:
                    raise UnresolvedDependencyException(task)

            results = pool.map(lambda x: x.run(), scheduled)

            # check that all tasks have produced results
            for task in scheduled:
                for target in to_list(task.target()):
                    if not target.exists():
                        raise NoResultException(task, target=target)

            for task, result in zip(scheduled, results):
                all_results[task] = result

    print(f"end => {datetime.now()} | elapsed: {(datetime.now() - start).total_seconds():.2f} seconds")
    return all_results



import json
import ast
from .base import Task
from .schedule import run

from argparse import ArgumentParser
import argparse
import schedule
import time


def import_task(name: str):
    components = name.split('.')
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


weekdays = ['monday', 'tuesday', 'wednesday', 'tursday', 'friday', 'saturday', 'sunday']
units = ['hour', 'minute', 'second', 'day', 'month', 'year']


def main():
    p = ArgumentParser()
    p.add_argument('--task', required=True)
    p.add_argument('--every', choices=weekdays + units + ["one_time"], default='one_time')
    p.add_argument('--count', type=int, default=None)
    p.add_argument('--to', type=int, default=None)
    p.add_argument('--at', help='time of the day, if not set, executed at 0:00')
    p.add_argument('args', nargs=argparse.REMAINDER)
    args = p.parse_args()

    task_class = args.task

    def parse_arg(x: str):

        for f in [ast.literal_eval, json.loads]:
            try:
                t = f(x)
                if isinstance(t, str):
                    raise TypeError('invalid type')
            except:
                pass

        return x

    task_args = list(map(parse_arg, args.args))
    assert (len(task_args) % 2 == 0)

    items = [(
        task_args[i], task_args[i + 1]
    ) for i in range(0, len(task_args), 2)]

    task_args = dict(items)

    TaskClass = import_task(task_class)

    print(f"restored task {TaskClass}")
    task: Task = TaskClass(**task_args)

    if args.every == 'one_time':
        results = run(task)
        print(f"results: {results}")
    else:
        sched = schedule
        if args.count is not None:
            sched = sched.every(args.count)
        else:
            sched = sched.every()

        if args.to is not None:
            sched = sched.to(args.to)

        sched = getattr(sched, args.every + "s")
        if args.at:
            sched = sched.at(args.at)

        sched.do(lambda: run(task))
        print("jobs: ", schedule.get_jobs())

        while True:

            # Checks whether a scheduled task
            # is pending to run or not
            schedule.run_pending()
            time.sleep(1)


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import List, Union, Optional
import logging
import io

try:
    import paramiko
    from paramiko.client import SSHClient, AutoAddPolicy
except:
    print("SSH Tasks cannot be used because 'paramiko' is not installed. <pip install paramiko>")


def get_client(ip: str, username: str, pem: Optional[Union[str, Path]] = None):

    client = SSHClient()
    client.load_system_host_keys()
    client.set_missing_host_key_policy(AutoAddPolicy())

    if pem is not None:
        if isinstance(pem, Path):
            logging.info(f"connecting via pem {pem}")
            private_key = paramiko.RSAKey.from_private_key(io.StringIO(pem))
            client.connect(ip, username=username, pkey=private_key)
        else:
            logging.info(f"connecting via pem path={pem}")
            client.connect(ip, username='ubuntu', key_filename=pem)
    else:
        client.connect(ip, username=username)

    return client


def upload_file(client: SSHClient, local_path: Union[str, Path], remote_path: Union[str, Path]):
    ftp_client = client.open_sftp()

    ftp_client.put(
        str(Path(local_path).absolute()),
        str(Path(remote_path).absolute())
    )

    ftp_client.close()


def download_file(client: SSHClient, remote_path: Union[str, Path], local_path: Union[str, Path]):
    ftp_client = client.open_sftp()

    ftp_client.put(
        str(Path(remote_path).absolute()),
        str(Path(local_path).absolute())
    )

    ftp_client.close()
